import numpy as np

from balderhub.waveform.lib.utils.waveforms.base_periodic_waveform import BasePeriodicWaveform


class CardiacWaveform(BasePeriodicWaveform):
    """simple cardiac waveform"""

    @property
    def points_per_period(self) -> int:
        """
        :return: returns the number of points the raw-data should have
        """
        return 16384

    @property
    def data(self) -> np.ndarray:
        # Phase von 0 bis fast 1 (endpoint=False → nahtlos wiederholbar)
        phase = np.linspace(0, 1, self.points_per_period, endpoint=False)

        signal = np.zeros(self.points_per_period, dtype=float)

        # P-Welle (Vorhof)
        signal += 0.25 * np.exp(-((phase - 0.18) ** 2) / (2 * 0.045 ** 2))

        # QRS complex (ventricle – the distinctive spike)
        qrs = 0.38
        signal += -0.15 * np.exp(-((phase - (qrs - 0.025)) ** 2) / (2 * 0.018 ** 2))  # Q
        signal += 1.00 * np.exp(-((phase - qrs) ** 2) / (2 * 0.012 ** 2))  # R (Hauptpeak)
        signal += -0.35 * np.exp(-((phase - (qrs + 0.035)) ** 2) / (2 * 0.022 ** 2))  # S

        # T wave (ventricular recovery)
        signal += 0.40 * np.exp(-((phase - 0.68) ** 2) / (2 * 0.085 ** 2))

        return signal



import numpy as np

from balderhub.waveform.lib.utils.waveforms.base_periodic_waveform import BasePeriodicWaveform

class DCWaveform(BasePeriodicWaveform):
    """simple DC Waveform"""

    @property
    def data(self) -> np.ndarray:
        return np.array([0, 0])



from __future__ import annotations

import pathlib
from abc import ABC, abstractmethod
from typing import TypeVar

import matplotlib.pyplot as plt

import numpy as np


AbstractWaveformTypeT = TypeVar('AbstractWaveformTypeT', bound='AbstractWaveform')


class AbstractWaveform(ABC):
    """
    Abstract base class describing a waveform
    """

    @property
    @abstractmethod
    def data(self) -> np.ndarray:
        """needs to return a list with values between 0 and 1"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def delta_time_sec(self) -> float:
        """
        :return: delta time between two data points
        """

    def _validate_data(self):
        data = self.data
        if max(data) > 1 or min(data) < -1:
            raise ValueError(f'data values needs to be between -1 and 1, but is between {min(data)} and {max(data)}')

    @abstractmethod
    def get_data_in_volts(self) -> np.ndarray:
        """
        This method converts the raw data according to the waveform parameters (like amplitude, phase, ..) into a
        numpy array that descibes the exact volt values.
        :return: a numpy array with voltage values
        """

    @abstractmethod
    def get_resampled_version(self, interval_sec: float) -> AbstractWaveformTypeT:
        """
        This method resamples the internal raw data to ensure that the interval_sec is the exact time between every
        step.

        :param interval_sec: the interval between every single data point
        :return: a new waveform instance with the resampled data points as raw data
        """

    @abstractmethod
    def compare(self, with_waveform: AbstractWaveform, max_allowed_rmse: float = 0.01) -> bool:
        """
        This method compares to waveforms with each other. If one of this waveforms is a PERIODIC and the other one is
        a NON-PERIODIC it will automatically try to convert the NON-PERIODIC waveform to a PERIODIC waveform.

        :param with_waveform: the other waveform to compare with
        :param max_allowed_rmse: maximal allowed RMSE value to classify both waveforms as identically
        :return: True if both waveforms are identical, otherwise False
        """

    @abstractmethod
    def _prepare_matplot(self):
        pass

    def save_plot(self, filepath: pathlib.Path) -> None:
        """
        This method creates a matplotlib plot and saves it at the given filepath.
        :param filepath: the filepath to save the plot to
        """
        self._prepare_matplot()
        plt.savefig(filepath)

    def show_plot(self) -> None:
        """
        This method shows a matplotlib plot of the waveform.
        """
        self._prepare_matplot()
        plt.show()



from __future__ import annotations
from abc import ABC
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import resample, correlate, find_peaks, detrend

from balderhub.waveform.lib.utils.waveforms.base_periodic_waveform import BasePeriodicWaveform
from balderhub.waveform.lib.utils.waveforms.custom_periodic_waveform import CustomPeriodicWaveform
from balderhub.waveform.lib.utils.waveforms.abstract_waveform import AbstractWaveform


class BaseNonPeriodicWaveform(AbstractWaveform, ABC):
    """
    Base abstract class for describing a NON-PERIODIC waveform
    """

    def __init__(
            self,
            multiplier_amplitude_volt: float,
            delta_time_sec: float,
            offset_vdc: float = 0
    ):
        """

        :param multiplier_amplitude_volt: full amplitude for the whole range - the full range is
                                          [-<multiplier_amplitude_volt>;+<multiplier_amplitude_volt>] (internal data
                                          will be multiplied with this value)
        :param delta_time_sec: the delta time between two data points
        :param offset_vdc: the offset in volts the signal is moved
        """
        self._multiplier_amplitude_volt = multiplier_amplitude_volt
        self._delta_time_sec = delta_time_sec
        self._offset_vdc = offset_vdc

    @property
    def multiplier_amplitude_volt(self) -> float:
        """
        :return: multiplier the raw data needs to be multiplied to define the voltage
        """
        return self._multiplier_amplitude_volt

    @property
    def delta_time_sec(self) -> float:
        return self._delta_time_sec

    @property
    def offset_vdc(self) -> float:
        """
        :return: additional offset that will be added to the signal
        """
        return self._offset_vdc

    @property
    def total_time_sec(self) -> float:
        """
        :return: returns the total time of the signal in seconds
        """
        return (len(self.data) - 1) * self._delta_time_sec

    def get_data_in_volts(self) -> np.ndarray:
        return self.data * float(self._multiplier_amplitude_volt) + self._offset_vdc

    def get_resampled_version(self, interval_sec: float) -> BaseNonPeriodicWaveform:
        # pylint: disable-next=import-outside-toplevel
        from balderhub.waveform.lib.utils.waveforms.custom_non_periodic_waveform import CustomNonPeriodicWaveform

        new_data = resample(self.data, int(len(self.data) * float(self.delta_time_sec / interval_sec)))

        # make sure that it is in expected range
        new_data = np.round(new_data, decimals=6).astype(np.float64)

        return CustomNonPeriodicWaveform(
            data=new_data,
            multiplier_amplitude_volt=self._multiplier_amplitude_volt,
            delta_time_sec=interval_sec,
            offset_vdc=self._offset_vdc
        )

    def get_periodic_equivalent_waveform(self) -> CustomPeriodicWaveform:
        """
        This method converts the NON-PERIODIC waveform into a PERIODIC waveform. It uses correlation to determine the
        periodic pattern within the signal.

        It raises a ValueError in case it does not detect a valid periodic pattern within the data.

        :return: a new PERIOD waveform that describes the periodic pattern within this signal
        """
        # TODO rework
        self_data = self.data
        # do autocorrelation and try to find periodic signal
        signal_detrended = detrend(self_data)
        auto_corr = correlate(signal_detrended, signal_detrended, mode='full')
        auto_corr = auto_corr[len(auto_corr) // 2:]  # only use positives

        # first try to find them with very high height, then go down
        for height_percent, min_distance in ((0.8, 50), (0.6, 100), (0.4, 200)):
            peaks, _ = find_peaks(
                auto_corr,
                height=height_percent * np.max(auto_corr),
                distance=min(min_distance, len(self_data) // 10)) # TODO
            if len(peaks) > 2:
                break

        if len(peaks) < 2:
            raise ValueError('no peaks found')

        periodic_samples = peaks[1] - peaks[0]  # TODO should we consider other peaks too?

        # extract periodic part
        n_complete_cycles = len(self_data) // periodic_samples

        if n_complete_cycles < 2:
            raise ValueError("to less data to determine periodic signal")

        # Only take complete cycles and pack them into the matrix
        trimmed = self_data[:n_complete_cycles * periodic_samples]
        cycles = trimmed.reshape((n_complete_cycles, periodic_samples))

        # TODO calc the error between the different cycles (needs to be a specific value)

        # Average all cycles → “clean average cycle”
        mean_cycle = np.mean(cycles, axis=0)

        return CustomPeriodicWaveform(
            data=mean_cycle,
            frequency_hz=float(1 / (self._delta_time_sec * len(mean_cycle))),
            amplitude_vpp=float(self._multiplier_amplitude_volt * 2),
            offset_vdc=self._offset_vdc,
            phase=0
        )

    def compare(
            self,
            with_waveform: Union[BasePeriodicWaveform, BaseNonPeriodicWaveform],
            max_allowed_rmse=0.01
    ) -> bool:

        if isinstance(with_waveform, CustomPeriodicWaveform):
            # one is periodic, the other is non-periodic
            self_in_periodic = self.get_periodic_equivalent_waveform()
            return self_in_periodic.compare(with_waveform, max_allowed_rmse=max_allowed_rmse)

        # both are non-periodic -> compare each other directly
        common_delta_time = max(self.delta_time_sec, with_waveform.delta_time_sec)
        self_data = self.get_resampled_version(common_delta_time).get_data_in_volts()
        other_data = with_waveform.get_resampled_version(common_delta_time).get_data_in_volts()
        if len(self_data) != len(other_data):
            return False
        rmse = np.sqrt(np.mean((self_data - other_data)**2))
        return rmse < max_allowed_rmse

    def _prepare_matplot(self):
        plt.figure(figsize=(10, 4))  # TODO

        time_points = np.arange(0, self.total_time_sec + self._delta_time_sec, self._delta_time_sec)

        plt.plot(time_points, self.get_data_in_volts())
        plt.title(f"{self.__class__.__name__}")
        plt.xlabel("Time [s]")
        plt.ylabel("Amplitude [V]")
        plt.grid(True)



from . import common

from .abstract_waveform import AbstractWaveform
from .base_non_periodic_waveform import BaseNonPeriodicWaveform
from .base_periodic_waveform import BasePeriodicWaveform
from .custom_non_periodic_waveform import CustomNonPeriodicWaveform
from .custom_periodic_waveform import CustomPeriodicWaveform

__all__ = [
    'AbstractWaveform',
    'BaseNonPeriodicWaveform',
    'BasePeriodicWaveform',
    'CustomNonPeriodicWaveform',
    'CustomPeriodicWaveform'
]



import numpy as np

from balderhub.waveform.lib.utils.waveforms.base_periodic_waveform import BasePeriodicWaveform


class SineWaveform(BasePeriodicWaveform):
    """simple sine waveform"""

    @property
    def points_per_period(self) -> int:
        """
        :return: returns the number of points the raw-data should have
        """
        return 16384

    @property
    def data(self) -> np.ndarray:
        return np.sin(np.linspace(0, 2 * np.pi, self.points_per_period))



from .cardiac_waveform import CardiacWaveform
from .cosine_waveform import CosineWaveform
from .dc_waveform import DCWaveform
from .sine_waveform import SineWaveform


__all__ = [
    'CardiacWaveform',
    'CosineWaveform',
    'SineWaveform',
    'DCWaveform',
]



import numpy as np

from .base_periodic_waveform import BasePeriodicWaveform


class CustomPeriodicWaveform(BasePeriodicWaveform):
    """
    Custom individual PERIODOC waveform that expects its data in the constructor
    """

    def __init__(
            self,
            data: np.ndarray,
            frequency_hz: float,
            amplitude_vpp: float,
            offset_vdc: float = 0,
            phase: float = 0
    ) -> None:
        super().__init__(
            frequency_hz=frequency_hz,
            amplitude_vpp=amplitude_vpp,
            offset_vdc=offset_vdc,
            phase=phase
        )
        self._data = data
        self._validate_data()

    @property
    def data(self) -> np.ndarray:
        return self._data



from __future__ import annotations
from abc import ABC
from typing import Union, TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from numpy.ma.core import correlate
from scipy.signal import resample

from balderhub.waveform.lib.utils.waveforms.abstract_waveform import AbstractWaveform

if TYPE_CHECKING:
    from balderhub.waveform.lib.utils.waveforms import BaseNonPeriodicWaveform


class BasePeriodicWaveform(AbstractWaveform, ABC):
    """
    Base abstract class for describing a PERIODIC waveform
    """
    def __init__(
            self,
            frequency_hz: float,
            amplitude_vpp: float,
            offset_vdc: float = 0,
            phase: float = 0
    ):
        self._frequency_hz = frequency_hz
        self._amplitude_vpp = amplitude_vpp
        self._offset_vdc = offset_vdc
        self._phase = phase % (2*np.pi)

    @property
    def frequency_hz(self) -> float:
        """
        :return: the frequency that should be applied to this waveform
        """
        return self._frequency_hz

    @property
    def delta_time_sec(self) -> float:
        return (1 / self.frequency_hz) / len(self.data)

    @property
    def amplitude_vpp(self) -> float:
        """
        :return: the amplitude value the raw data will be multiplied with
        """
        return self._amplitude_vpp

    @property
    def offset_vdc(self) -> float:
        """
        :return: the signal offset the raw data will be added to
        """
        return self._offset_vdc

    @property
    def phase(self) -> float:
        """
        :return: the phase of the signal
        """
        return self._phase

    def get_data_in_volts(self) -> np.ndarray:
        split_at = int(len(self.data) * (self.phase % (2 * np.pi)) / (2 * np.pi))
        phase_cleaned_raw_data = np.concatenate((self.data[split_at:], self.data[:split_at]))

        return phase_cleaned_raw_data * float(self._amplitude_vpp) / 2 + self._offset_vdc

    def get_resampled_version(self, interval_sec: float) -> BasePeriodicWaveform:
        # pylint: disable-next=import-outside-toplevel
        from balderhub.waveform.lib.utils.waveforms.custom_periodic_waveform import CustomPeriodicWaveform

        new_data_point_cnt = int(len(self.data) * float(self.delta_time_sec / interval_sec))
        if new_data_point_cnt == 0:
            raise ValueError('can not resample data, because no data point would be there with the new interval time')
        new_data = resample(self.data, new_data_point_cnt)

        # make sure that it is in expected range
        new_data = np.round(new_data, decimals=6).astype(np.float64)

        return CustomPeriodicWaveform(
            data=new_data,
            frequency_hz=self.frequency_hz,
            amplitude_vpp=self.amplitude_vpp,
            offset_vdc=self.offset_vdc,
            phase=self.phase
        )

    def get_phase_difference_to(
            self,
            other_periodic_waveform: BasePeriodicWaveform,
            max_allowed_rmse: float = 0.01,
    ) -> Union[float, None]:
        """
        This method gets the phase difference between two periodic waveforms.

        :param other_periodic_waveform: the other waveform
        :return: the phase between [0,2*pi] from this waveform to the other waveform or None, if both waveforms are
                 different
        """
        # pylint: disable-next=import-outside-toplevel
        from balderhub.waveform.lib.utils.waveforms import CustomPeriodicWaveform

        if self.amplitude_vpp != other_periodic_waveform.amplitude_vpp:
            return False
        if self.frequency_hz != other_periodic_waveform.frequency_hz:
            return False
        common_delta_time = max(self.delta_time_sec, other_periodic_waveform.delta_time_sec)
        self_signal = self.get_resampled_version(common_delta_time).get_data_in_volts()
        other_signal = other_periodic_waveform.get_resampled_version(common_delta_time).get_data_in_volts()

        # double one of the curves
        self_signal = np.tile(self_signal, 2)
        # determine the position with the best correlation
        auto_corr = correlate(self_signal, other_signal, mode='valid')
        best_pos = np.argmax(auto_corr)
        self_signal = self_signal[best_pos:best_pos + len(self.data)]
        if len(self_signal) < len(other_signal):
            return False
        phase = 2 * np.pi * (best_pos/len(self.data))
        new_modified_waveform = CustomPeriodicWaveform(
            data=other_periodic_waveform.data,
            frequency_hz=other_periodic_waveform.frequency_hz,
            amplitude_vpp=other_periodic_waveform.amplitude_vpp,
            offset_vdc=other_periodic_waveform.offset_vdc,
            phase=other_periodic_waveform.phase - phase
        )

        if self.compare(new_modified_waveform, max_allowed_rmse, ignore_phase=False):
            return phase
        return None

    def compare(
            self,
            with_waveform: Union[BasePeriodicWaveform, BaseNonPeriodicWaveform],
            max_allowed_rmse: float = 0.01,
            ignore_phase: bool = False
    ) -> bool:
        """
        This method compares to waveforms with each other. If one of the waveforms is a PERIODIC and the other one is
        a NON-PERIODIC it will automatically try to convert the NON-PERIODIC waveform to a PERIODIC waveform.

        If ``ignore_phase`` is True, it will ignore the position of the waveform and attempts to find the same
        waveform by shifting it. For example, in a cosine and sine comparison where ``ignore_phase=True``, True
        would also be returned in the compare, since the signals are identical but with a phase shift of 90 degree.

        :param with_waveform: the other waveform to compare with
        :param max_allowed_rmse: maximal allowed RMSE value to classify both waveforms as identically
        :return: True if both waveforms are identical, otherwise False
        """
        # TODO improve
        # pylint: disable-next=import-outside-toplevel
        from balderhub.waveform.lib.utils.waveforms.custom_non_periodic_waveform import CustomNonPeriodicWaveform

        if isinstance(with_waveform, CustomNonPeriodicWaveform):
            # one is periodic, the other is non-periodic
            with_waveform_periodic = with_waveform.get_periodic_equivalent_waveform()
            return with_waveform_periodic.compare(
                with_waveform,
                max_allowed_rmse=max_allowed_rmse,
                ignore_phase=ignore_phase
            )
        # have the same type -> compare each other directly
        if self.frequency_hz != with_waveform.frequency_hz:
            return False

        if ignore_phase:
            return self.get_phase_difference_to(with_waveform, max_allowed_rmse=max_allowed_rmse) is not None

        common_delta_time = max(self.delta_time_sec, with_waveform.delta_time_sec)
        self_signal = self.get_resampled_version(common_delta_time).get_data_in_volts()
        other_signal = with_waveform.get_resampled_version(common_delta_time).get_data_in_volts()

        rmse = np.sqrt(np.mean((self_signal - other_signal) ** 2))
        return rmse < max_allowed_rmse

    def _prepare_matplot(self):
        # TODO
        plt.figure(figsize=(10, 4))  # TODO

        delta_time_sec = (1 / self.frequency_hz) / len(self.data)

        time_points = np.arange(0, 1 / self.frequency_hz, delta_time_sec)

        plt.plot(time_points, self.get_data_in_volts())
        plt.title(f"{self.__class__.__name__}")
        plt.xlabel("Time [s]")
        plt.ylabel("Amplitude [V]")
        plt.grid(True)




import balder
import numpy as np

from balderhub.unit.scenarios import ScenarioUnit

from balderhub.waveform.lib.utils import waveforms


class ScenarioCheckWaveformMethods(ScenarioUnit):

    @balder.fixture('scenario')
    def sinus(self):
        return waveforms.common.SineWaveform(
            frequency_hz=5,
            amplitude_vpp=2
        )

    @balder.fixture('scenario')
    def cosinus(self):
        return waveforms.common.CosineWaveform(
            frequency_hz=5,
            amplitude_vpp=2
        )

    def test_convert_non_periodic_to_periodic(self, sinus, cosinus):

        non_periodic = waveforms.CustomNonPeriodicWaveform(
            np.sin(np.linspace(0, 10 * np.pi, 10000)),
            delta_time_sec=0.0001,
            multiplier_amplitude_volt=1
        )

        periodic_equivalent = non_periodic.get_periodic_equivalent_waveform()
        assert float(periodic_equivalent.frequency_hz) == 5
        assert periodic_equivalent.compare(sinus, ignore_phase=True)
        assert periodic_equivalent.compare(cosinus, ignore_phase=True)

    def test_compare(self, sinus):
        assert sinus.compare(sinus), f"did not return True when comparing the same object"

    def test_compare_with_ignore_phase(self, sinus, cosinus):
        assert not sinus.compare(cosinus)
        assert sinus.compare(cosinus, ignore_phase=True)



import numpy as np

from .base_non_periodic_waveform import BaseNonPeriodicWaveform


class CustomNonPeriodicWaveform(BaseNonPeriodicWaveform):
    """
    Custom individual NON-PERIODOC waveform that expects its data in the constructor
    """

    def __init__(
            self,
            data: np.ndarray,
            multiplier_amplitude_volt: float,
            delta_time_sec: float,
            offset_vdc: float = 0
    ) -> None:
        super().__init__(
            multiplier_amplitude_volt=multiplier_amplitude_volt,
            delta_time_sec=delta_time_sec,
            offset_vdc=offset_vdc,
        )
        self._data = data
        self._validate_data()

    @property
    def data(self) -> np.ndarray:
        return self._data
